

#pragma once

#include <atlmisc.h>
#include <boost/call_traits.hpp>
#include <boost/concept_check.hpp>
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/mpl/apply.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/eval_if.hpp>
#include <boost/mpl/is_sequence.hpp>
#include <boost/mpl/placeholders.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/spirit/fusion/sequence/tuple.hpp>
#include <boost/spirit/fusion/sequence/get.hpp>
#include <cassert>

#include "detail/gen_process_statement.hpp"
#include "versional.hpp"
#include "message_processor.hpp"

namespace ketchup {
	
	namespace detail {

	struct update_cmd_ui_base
	{
		bool b_cmd_ui_handled;
		bool is_cmd_ui_handled() const { return b_cmd_ui_handled; }
		void set_cmd_ui_handled(bool b_) { b_cmd_ui_handled = b_; }
	};

	} // namespace detail

// cmd_ui types

struct cmd_ui
{
	virtual void enable(bool bOn = true) = 0;
	virtual void set_check(int nCheck = 1) = 0;
	virtual void set_radio(bool bOn = true) = 0;
	virtual void set_text(LPCTSTR) = 0;
	virtual ~cmd_ui() { }
	virtual bool is_popup() { return false; } // popup menu can't have id...
};

struct menu_cmd_ui : cmd_ui
{
	CMenuHandle m_menu;
	UINT m_nIndex;	// for the performance
	bool m_bPopup;

	explicit menu_cmd_ui(HMENU hMenu, UINT nIndex, bool bPopup)
		: m_menu(hMenu), m_nIndex(nIndex), m_bPopup(bPopup)
	{
		ATLASSERT(m_menu.IsMenu());
	}

	virtual void enable(bool bOn = true)
	{
		m_menu.EnableMenuItem(m_nIndex, MF_BYPOSITION |
				(bOn ? MF_ENABLED : (MF_DISABLED | MF_GRAYED)));
	}

	virtual void set_check(int nCheck = 1)
	{
		m_menu.CheckMenuItem(m_nIndex, MF_BYPOSITION |
			(nCheck ? MF_CHECKED : MF_UNCHECKED));
	}

	virtual void set_radio(bool bOn = true)
	{
		set_check(bOn ? 1 : 0); // not yet
	}

	virtual void set_text(LPCTSTR lpszText)
	{
		assert(lpszText != NULL);

		// get current menu state so it doesn't change
		UINT nState = m_menu.GetMenuState(m_nIndex, MF_BYPOSITION);
		nState &= ~(MF_BITMAP|MF_OWNERDRAW|MF_SEPARATOR);

		// set menu text
		versional<MENUITEMINFO> mii;
		mii.fMask = MIIM_ID;
    ATLVERIFY(m_menu.GetMenuItemInfo(m_nIndex, TRUE, &mii));
		ATLVERIFY(m_menu.ModifyMenu(m_nIndex, MF_BYPOSITION |
			MF_STRING | nState, mii.wID, lpszText));
	}

	virtual bool is_popup() { return m_bPopup; }
};

struct tool_cmd_ui : cmd_ui
{
	CToolBarCtrl m_wndToolBar;
	UINT m_nID;

	explicit tool_cmd_ui(HWND hWnd, UINT nID)
		: m_wndToolBar(hWnd), m_nID(nID)
	{
		ATLASSERT(m_wndToolBar.IsWindow());
	}

	// overrides
	virtual void enable(bool bOn = true)
	{
		// See: See: MFC7::CToolCmdUI::Enable
		UINT nOldState = m_wndToolBar.GetState(m_nID);
		UINT nNewState = nOldState;
		assert(nNewState != -1);
		if (!bOn) {
			nNewState &= ~TBSTATE_ENABLED;
			// WINBUG: If a button is currently pressed and then is disabled
			// COMCTL32.DLL does not unpress the button, even after the mouse
			// button goes up!  We work around this bug by forcing TBBS_PRESSED
			// off when a button is disabled.
			nNewState &= ~TBSTATE_CHECKED;
		}
		else {
			nNewState |= TBSTATE_ENABLED;
		}

		ATLVERIFY(m_wndToolBar.SetState(m_nID, nNewState));
	}

	virtual void set_check(int nCheck = 1)
	{
		assert(nCheck >= 0 && nCheck <= 2); // 0=>off, 1=>on, 2=>indeterminate

		versional<TBBUTTONINFO> tbb;
		tbb.dwMask = TBIF_STATE | TBIF_STYLE;
		ATLVERIFY( m_wndToolBar.GetButtonInfo(m_nID, &tbb) != -1 );

		// add check style
		tbb.fsStyle = tbb.fsStyle | TBSTYLE_CHECK;

		// reset state and...
		tbb.fsState = tbb.fsState &
					~(TBSTATE_CHECKED | TBSTATE_INDETERMINATE);

		if (nCheck == 1)
			tbb.fsState |= TBSTATE_CHECKED;
		else if (nCheck == 2)
			tbb.fsState |= TBSTATE_INDETERMINATE;

		ATLVERIFY(m_wndToolBar.SetButtonInfo(m_nID, &tbb));
	}

	virtual void set_radio(bool bOn = true)
	{
		set_check(bOn); // not yet
	}

	virtual void set_text(LPCTSTR)
	{
		// ignore it
	}
};

///////////////////////////////////////////////////////////////////////////////
// update_cmd_ui
//
template< class Derived >
struct update_cmd_ui : detail::update_cmd_ui_base, message_processor<update_cmd_ui, true>
{
	typedef boost::fusion::tuple<Derived&, cmd_ui&, UINT> process_update_cmd_ui_args_type;
	typedef typename boost::call_traits<process_update_cmd_ui_args_type>::reference process_update_cmd_ui_args_ref;
	typedef typename boost::call_traits<process_update_cmd_ui_args_type>::param_type process_update_cmd_ui_args_param_type;

	// shortcuts
	static Derived& get_derived(process_update_cmd_ui_args_param_type xs)
	{
		return boost::fusion::get<0>(xs);
	}

	static cmd_ui& get_cmd_ui(process_update_cmd_ui_args_param_type xs)
	{
		return boost::fusion::get<1>(xs);
	}

	static UINT get_nID(process_update_cmd_ui_args_param_type xs)
	{
		return boost::fusion::get<2>(xs);
	}

	template< class UpdateCmdUIMap >
	bool process_update_cmd_ui(cmd_ui& ui, UINT nID)
	{
		Derived& derived = *static_cast<Derived*>(this);
		bool old = derived.is_cmd_ui_handled();
		typedef typename detail::gen_process_statement<UpdateCmdUIMap, process_update_cmd_ui_args_ref>::type statement;
		process_update_cmd_ui_args_type xs(derived, ui, nID);
		bool ret = statement::process(xs);
		derived.set_cmd_ui_handled(old);
		return ret;
	}
	
	template< class UpdateCmdUIMap >
	void update_tool_bar_cmd_ui(HWND hWndToolBar)
	{
		// See: CToolBar::OnUpdateCmdUI at Vc7/atlmfc/src/mfc/bartool.cpp
		CToolBarCtrl wndToolBar(hWndToolBar);
		int nIndexMax = wndToolBar.GetButtonCount();
		for (int i = 0; i < nIndexMax; ++i)
		{
			// get buttons state
			TBBUTTON button;
			wndToolBar.GetButton(i, &button);
			UINT nID = button.idCommand;
			tool_cmd_ui ui(wndToolBar, nID);

			// ignore separators
			if (!(button.fsStyle & TBSTYLE_SEP))
				process_update_cmd_ui<UpdateCmdUIMap>(ui, nID);
		}
	}

	template< class UpdateCmdUIMap >
	void update_menu_cmd_ui(CMenuHandle menu)
	{
		// See: MFC7::CFrameWnd::OnInitMenuPopup
		int count = menu.GetMenuItemCount();
		for (int i = 0; i < count; ++i)
		{
			UINT nID = menu.GetMenuItemID(i);
			if (nID == 0)
				continue; // menu separator or invalid cmd - ignore it

			if (nID == (UINT)-1)
			{
				// possibly a popup menu, route to first item of that popup
				CMenuHandle sub_menu = menu.GetSubMenu(i);
				if (sub_menu.m_hMenu == NULL ||
					(nID = sub_menu.GetMenuItemID(0)) == 0 ||
					nID == (UINT)-1)
				{
					continue;       // first item of popup can't be routed to
				}
				menu_cmd_ui ui(menu, i, true);
				process_update_cmd_ui<UpdateCmdUIMap>(ui, nID); // popups
			}
			else
			{
				// normal menu item
				// Auto enable/disable if frame window has 's_bAutoMenuEnable'
				//    set and command is _not_ a system command.
				menu_cmd_ui ui(menu, i, false);
				process_update_cmd_ui<UpdateCmdUIMap>(ui, nID);
			}

			// adjust for menu deletions and additions
			int new_count = menu.GetMenuItemCount();
			if (new_count < count)
			{
				i -= (count - new_count);
				while (i < new_count &&
					menu.GetMenuItemID(i) == nID)
				{
					++i;
				}
			}
			count = new_count;
		} // for
	}

	// update_command_ui
	template< UINT id, void (Derived::*func)(cmd_ui&) >
	struct update_command_ui
	{
		static bool process(process_update_cmd_ui_args_ref xs)
		{
			if (get_nID(xs) == id) {
				get_derived(xs).set_cmd_ui_handled(true);
				(get_derived(xs).*func)(get_cmd_ui(xs));
				if (get_derived(xs).is_cmd_ui_handled())
					return true;
			}
			return false;
		}
	};
	template< UINT id, void (Derived::*func)(cmd_ui&) >
	struct empty_command_ui
	{
		static bool process(process_update_cmd_ui_args_ref xs)
		{
			return false;
		}
	};

	void OnInitMenuPopup(HMENU hMenu, UINT, BOOL)
	{
		set_msg_handled(false);
		update_menu_cmd_ui<Derived::update_cmd_ui_map>(hMenu);
	}

	struct message_map : boost::mpl::vector<
		msg_wm_initmenupopup<&update_cmd_ui::OnInitMenuPopup>
	> { };

};

} // namespace ketchup

